#ifndef __STRASSEN_P_H__
#define __STRASSEN_P_H__


#include "strassen_kernels.h"


// P1 = (A11 + A22) * (B11 + B22)
static inline __attribute__((always_inline)) void strassen_P1(int b, int c, int d, int m, int k, int n, int bA, int bB, int dimA, int dimB, double *A11, double *A22, double *B11, 
	double *B22, double *P1, double *S, unsigned long *scratchs) {

	unsigned long lscratchs = *scratchs;

	double *F = S + lscratchs;
	lscratchs += dimA;
	double *G = S + lscratchs;
	lscratchs += dimB;

	int i;  
	for ( i = 0; i < dimA; i += bA ) {
		basic_madd(bA, &A11[i], &A22[i], &F[i]); 
	}
	for ( i = 0; i < dimB; i += bB ) {
		basic_madd(bB, &B11[i], &B22[i], &G[i]); 
	}

	*scratchs = lscratchs;

	strassen_mm(b, c, d, m, k, n, F, G, P1, S, scratchs);	
}


// P2 = (A21 + A22) * B11
static inline __attribute__((always_inline)) void strassen_P25(int b, int c, int d, int m, int k, int n, int bA, int bB, int dimA, int dimB, double *A21, double *A22, double *B11, 
	double *P25, double *S, unsigned long *scratchs) {
	unsigned long lscratchs = *scratchs;

	double *F = S + lscratchs;
	lscratchs += dimA;

	int i = 0;
	for ( i = 0; i < dimA; i += bA ) {
		basic_madd(bA, &A21[i], &A22[i], &F[i]); 
	}

	*scratchs = lscratchs;

	strassen_mm(b, c, d, m, k, n, F, B11, P25, S, scratchs);	
}


// P3 = A11 * (B12 - B22) 
static inline __attribute__((always_inline)) void strassen_P34(int b, int c, int d, int m, int k, int n, int bA, int bB, int dimA, int dimB, double *A11, double *B12, 
	double *B22, double *P34, double *S, unsigned long *scratchs) {
	unsigned long lscratchs = *scratchs;

	double *F = S + lscratchs;
	lscratchs += dimB;

	int i = 0;
	for ( i = 0; i < dimB; i += bB ) {
		basic_msub(bB, &B12[i], &B22[i], &F[i]); 
	}

	*scratchs = lscratchs;

	strassen_mm(b, c, d, m, k, n, A11, F, P34, S, scratchs);	
}


// P3 = A11 * (B12 - B22) 
static inline __attribute__((always_inline)) void strassen_P3(int b, int c, int d, int m, int k, int n, int bA, int bB, int dimA, int dimB, double *A11, double *B12, 
	double *B22, double *P34, double *S, unsigned long *scratchs) {
	//unsigned long lscratchs = *scratchs;

	//double *F = S + lscratchs;
	//lscratchs += dimB;

	int i = 0;
	for ( i = 0; i < dimB; i += bB ) {
		basic_msub2(bB, &B22[i], &B12[i]);
	}

	//*scratchs = lscratchs;

	strassen_mm(b, c, d, m, k, n, A11, B12, P34, S, scratchs);	
}


// P6 = (A21 - A11) * (B11 + B12)
static inline __attribute__((always_inline)) void strassen_P67(int b, int c, int d, int m, int k, int n, int bA, int bB, int dimA, int dimB, double *A21, double *A11, double *B11, 
	double *B12, double *P67, double *S, unsigned long *scratchs) {
	unsigned long lscratchs = *scratchs;

	double *F = S + lscratchs;
	lscratchs += dimA;
	double *G = S + lscratchs;
	lscratchs += dimB;

	int i = 0;
	for ( i = 0; i < dimA; i += bA ) {
		basic_msub(bA, &A21[i], &A11[i], &F[i]); 
	}
	for ( i = 0; i < dimB; i += bB ) {
		basic_madd(bB, &B11[i], &B12[i], &G[i]); 
	}

	*scratchs = lscratchs;

	strassen_mm(b, c, d, m, k, n, F, G, P67, S, scratchs);	
}


#endif // __STRASSEN_P_H__
